package com.example.modacon;

import java.util.Map;
import java.util.LinkedHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.Arrays;

import com.dodo.tools.CommGeneric;

import android.app.Activity;
import android.widget.Toast;

// Modbus operations manager
public class ModbusManager {
    public static final int START_DELAY=50, REQ_DELAY=100, TEST_DELAY=600, RETRIES=4, RETRY_DELAY=200;

    public AtomicBoolean started    = new AtomicBoolean(true);  // operation thread started flag
    public AtomicBoolean polling    = new AtomicBoolean(true);  // periodic polling enabled flag
    public AtomicBoolean connecting = new AtomicBoolean(false); // connecting in progress flag
    public AtomicInteger slaveAddr  = new AtomicInteger(0);     // current Modbus slave address

    public Activity activity;
    public ModbusManager(Activity activity){
        this.activity = activity;
    }
    
    // requests queue
    private ConcurrentLinkedQueue<ModbusRequest>   requests = new ConcurrentLinkedQueue<ModbusRequest>();
    
    // replies
    public ConcurrentHashMap<Integer, ModbusReply> replies  = new ConcurrentHashMap<Integer, ModbusReply>();

    // available registers
    public LinkedHashMap<Integer,ModbusReg>            regs = new LinkedHashMap<Integer,ModbusReg>();
    
    // register groups
    public LinkedHashMap<Integer,Integer>         regGroups = new LinkedHashMap<Integer,Integer>();
    
    // started periodic polls
    public LinkedHashMap<Integer,ModbusPoll>          polls = new LinkedHashMap<Integer,ModbusPoll>();

    // current communication channel (USB/Bluetooth/TCP)
    public CommGeneric channel = null;
    
    public String debugS;

    // check if connection if ok
    public boolean connected(){ return started.get() && (channel != null) && channel.isOpen.get(); }

    // translation language and function
    public int lang = 0;
    public String tr(String s){
        String[] ts = s.trim().split("//");
        return ts[(lang < ts.length) ? lang : 0].replace("_"," ").trim();
    }

    // processing thread
    Thread pollThread = new Thread(new Runnable(){ public void run(){
        try{ Thread.sleep(START_DELAY);
        }catch(Exception e){};
        while(started.get()){
            // handle read/write requests in queue
            if(connected()){
                if(!requests.isEmpty()){
                    ModbusRequest req = requests.poll();
                    // retry request RETRIES times
                    int retries = RETRIES;
                    while(retries > 0){
                        byte[] reqBytes = req.getPacket();
                        if(req.values == null){ // 03 - read request
                            try{
                                byte repBytes[] = channel.read();
                                channel.write(reqBytes);
                                Thread.sleep(REQ_DELAY);
                                repBytes = channel.read();
                                int repRegs[] = ModbusRtu.parse03(req.addr, req.reg, req.nregs, repBytes);
                                if(repRegs.length > 0) retries = 0;
                                // add replies
                                replies.put(req.id, new ModbusReply(repRegs));
                                // update registers/UI if necessary
                                if(req.updateRegs){
                                    for(int i=0; i<repRegs.length; i++){
                                        if(regs.containsKey(req.reg + i))
                                            regs.get(req.reg + i).setValue(repRegs[i]);
                                    }
                                }
                            }catch(Exception e){}
                        }else{ // 16 - write request
                            try{
                                byte repBytes[] = channel.read();
                                channel.write(reqBytes);
                                Thread.sleep(REQ_DELAY);
                                repBytes = channel.read();
                                boolean ok = ModbusRtu.parse16(req.addr, req.reg, req.nregs, repBytes);
debugS = Arrays.toString(reqBytes) + " : " + Arrays.toString(repBytes) + " : " + ok + " : " + ModbusRtu.debugN + " : " + ModbusRtu.debugS;
//toast(debugS, true);
                                if(ok) retries = 0;
                                replies.put(req.id, new ModbusReply(ok ? new int[0] : null));
                            }catch(Exception e){}
                        }
                        if(retries > 0){ // pause between retries
                            try{ Thread.sleep(RETRY_DELAY);
                            }catch(Exception e){}
                            retries--;
                        }
                    };
                }
            }else if (!requests.isEmpty()) requests.clear();
            
            // remove old replies
            long timestamp = System.currentTimeMillis();
            for(Map.Entry<Integer,ModbusReply> entry : replies.entrySet())
                if(entry.getValue().timestamp <= timestamp) replies.remove(entry.getKey());
            
            // pause between requests
            try{ Thread.sleep(50);
            }catch(Exception e){}
        }
        // stop all polls
toast("stop...",true);
        polling.set(false);
        for(ModbusPoll poll : polls.values()) poll.stop();
        // close channel
        if(channel != null) channel.close();
        // clear requests/replies
        requests.clear();
        replies.clear();
    }});
    
    // connect to channel
    public void connect(CommGeneric channel){
        disconnect(); // disconnect from old channel
        if(channel.isOpen.get()){
            this.channel = channel;
            // start processing thread if necessary
            if(!pollThread.isAlive()) pollThread.start();
        }
    }
    // disconnect
    public void disconnect(){
        if(channel != null){
            channel.close();
            channel = null;
        }
    }
    // stop all
    public void stop(){
        disconnect();
        started.set(false);
    }
    
    public int regPrev=-2, regGroup;
    public ModbusReg lastReg;
    // add new register
    public ModbusManager addReg(ModbusReg reg){
        lastReg = reg;
        regs.put(reg.addr, reg);

        // add new/expand current register group
        regPrev++;
        if(reg.addr == regPrev){
            if(regGroups.containsKey(regGroup))
                regGroups.put(regGroup, regGroups.get(regGroup)+1);
        }else{
            regGroup = regPrev = reg.addr;
            regGroups.put(reg.addr,1);
        }
        
        return this;
    }
    public ModbusReg addReg(int _tab, String _type, String _name, String _descr, int _addr, int _digits){
        ModbusReg reg = new ModbusReg(_tab, _type, tr(_name),tr(_descr), _addr, _digits, -1,-1, this);
        addReg(reg);
        return reg;
    }
    public ModbusReg addReg(int _tab, String _type, String _name, String _descr, int _addr, int _digits, int _min, int _max){
        ModbusReg reg = new ModbusReg(_tab, _type,tr(_name),tr(_descr), _addr, _digits, _min,_max, this);
        addReg(reg);
        return reg;
    }

    // read registers
    public int readRegs(int addr, int reg, int nregs, boolean update){
        if(connected()){
            ModbusRequest req = new ModbusRequest(addr, reg, nregs, update);
            requests.add(req);
            return req.id;
        }else return -1;
    }
    public int readRegs(int reg, int nregs, boolean update){
        return readRegs(slaveAddr.get(), reg, nregs, update);
    }
    // write registers
    public int writeRegs(int addr, int reg, int[] regs, boolean update){
        if(connected()){
            ModbusRequest req = new ModbusRequest(addr, reg, regs, update);
//toast("write addr:"+addr + " reg:"+reg + " regs:"+Arrays.toString(regs), true);
            requests.add(req);
            return req.id;
        }else return -1;
    }
    public int writeRegs(int reg, int[] regs){
        return writeRegs(slaveAddr.get(), reg, regs, false);
    }
    // write single register
    public int writeReg(int reg, int value){
        return writeRegs(reg, new int[]{value});
    }
    // read group
    public void readGroup(int reg){
        if(regGroups.containsKey(reg))
            readRegs(reg, regGroups.get(reg), true);
    }
    // read page
    public void readPage(int page){
    }
    // read all
    public void readAll(){
        for(int reg : regGroups.keySet())
            readGroup(reg);
    }
    
    // add poll
    public ModbusPoll addPoll(int period, int reg, int regs){
        ModbusPoll poll = new ModbusPoll(this, period, reg, regs);
        polls.put(reg, poll);
        return poll;
    }

    // check for reply
    public boolean replied(int id){ return replies.containsKey(id); }
    // get reply
    public ModbusReply getReply(int id){
        return (replies.containsKey(id)) ? replies.get(id) : null;
    }
    // get/delete reply
    public ModbusReply getReply(int id, boolean del){
        if(replies.containsKey(id)){
            ModbusReply rep = replies.get(id);
            if(del) replies.remove(id);
            return rep;
        }else return null;
    }
    
    private String toastMsg;
    private boolean toastIsLong;
    public void toast(String msg, boolean isLong){ toastMsg=msg; toastIsLong=isLong; activity.runOnUiThread(new Runnable(){ public void run(){
        Toast.makeText(activity, toastMsg, toastIsLong ? Toast.LENGTH_LONG : Toast.LENGTH_SHORT).show();
    }}); };
    public void toast(String msg){ toast(msg,false); }
}
